/**
 * The default comparison function. Designed to be consistend with
 * `Array.prototype.sort()`.`
 */
export declare function defaultCmp(a: unknown, b: unknown): number;
/**
 * A sorted array. Stays sorted through insertion and deletion.
 */
export declare class SortedArray<T> {
    private root?;
    private compare;
    [Symbol.toStringTag]: string;
    /**
     * Constructs a new Sorted Array.
     *
     * @param compare The comparison function for the sorting. Compares two
     * elements a and b. If a is less than b, returns a negative number. If a is
     * greater than b returns a positive number. If a is equal to b returns zero.
     */
    constructor(compare?: (a: T, b: T) => number);
    private findAtIndex;
    /**
     * Removes the last value from the array and returns it. Optionally an index
     * to pop from can be specified.
     *
     * @param index The target index to pop from the array. Defaults to the last
     * element in the array. Must be within [0, arr.length)
     */
    pop(index?: number): T | undefined;
    /**
     * Inserts a value into the sorted array in sorted order.
     *
     * Note that if the comparison function returns zero on a value already in the
     * array, the inserted value will be "absorbed" into that value.
     *
     * For example:
     *
     * ```ts
     * const arr = SortedArray((a, b) => 0)
     *
     * arr.insert(5) // [ 5 ]
     * arr.insert(1000) // [ 5, 5 ]
     * ```
     *
     * The comparator returned 0 when called on 5 and 1000, so the 1000 value was
     * absorbed into the value 5.
     */
    insert(value: T): this;
    private indexRange;
    /**
     * Returns the largest index that the value could be inserted into.
     *
     * That is, `i`, such that for the sorted array `arr`, if `j` < `i`, `arr[j]`
     * ≤ `value` and if `j` ≥ `i`, `arr[j]` > `value`.
     */
    bisectRight(value: T): number;
    /**
     * Similar to [[bisectRight]], but returns the smallest index that the value
     * could be inserted to.
     */
    bisectLeft(value: T): number;
    /**
     * Returns the first index at which a given element can be found in the array,
     * or -1 if it is not present.
     *
     * @param value The value to locate in the array.
     * @param fromIdx The index to start the search at. If the index is greater
     * than or equal to the array's length, -1 is returned, which means the array
     * will not be searched. If the provided index value is a negative number, it
     * is taken as the offset from the end of the array. Note: if the provided
     * index is negative, the array is still searched from front to back. If the
     * provided index is 0, then the whole array will be searched. Default: 0
     * (entire array is searched).
     * @returns The first index of the element in the array; -1 if not found.
     */
    indexOf(value: T, fromIdx?: number): number;
    /**
     * Returns the last index at which a given element can be found in the array,
     * or -1 if it is not present. The array is searched backwards, starting at
     * fromIndex.
     *
     * @param value The value to locate in the array.
     * @param fromIndex The index at which to start searching backwards. Defaults
     * to the array's length minus one (arr.length - 1), i.e. the whole array will
     * be searched. If the index is greater than or equal to the length of the
     * array, the whole array will be searched. If negative, it is taken as the
     * offset from the end of the array. Note that even when the index is
     * negative, the array is still searched from back to front. If the calculated
     * index is less than 0, -1 is returned, i.e. the array will not be searched.
     * @returns The last index of the element in the array; -1 if not found.
     */
    lastIndexOf(value: T, fromIndex?: number): number;
    /**
     * Executes a provided function once for each array element.
     *
     * @param callbackfn The function to execute on each element.
     * @param thisArg The value to use as this when executing callback.
     */
    forEach(callbackfn: (value: T, index: number, array: this) => void, thisArg?: unknown): void;
    entries(): IterableIterator<[number, T]>;
    keys(): IterableIterator<number>;
    values(): IterableIterator<T>;
    /**
     * Returns the value at the given index in the sorted order. (0-indexed)
     *
     * If the index is invalid, returns undefined.
     *
     * @param idx The index to lookup
     */
    get(idx: number): T | undefined;
    /**
     * The number of distinct elements in the set.
     */
    get length(): number;
    /**
     * Iterates over the values in the set in sorted order.
     */
    [Symbol.iterator](): Generator<T>;
    /**
     * Creates a sorted array from an iterable.
     *
     * @param iterable The itereable to turn into a sorted array.
     * @param compare The compare method for the resulting SortedArray.
     * @returns A new sorted array whose elements are from the passed iterable.
     */
    static from<T>(iterable: Iterable<T>, compare?: (a: T, b: T) => number): SortedArray<T>;
}
