import { useWeightBalancedTree } from 'wbtree';
import { enumerate, range, repeat } from '@shlappas/itertools';

const { balanceLeft, balanceRight } = useWeightBalancedTree((a, b) => {
    var _a, _b;
    const aCount = a.data.valueCount;
    const bCount = b.data.valueCount;
    const blCount = (_b = (_a = b.left) === null || _a === void 0 ? void 0 : _a.data.valueCount) !== null && _b !== void 0 ? _b : 0;
    a.data.valueCount -= bCount - blCount;
    b.data.valueCount += aCount - bCount;
}, (b, c) => {
    var _a, _b;
    const bCount = b.data.valueCount;
    const cCount = c.data.valueCount;
    const brCount = (_b = (_a = b.right) === null || _a === void 0 ? void 0 : _a.data.valueCount) !== null && _b !== void 0 ? _b : 0;
    b.data.valueCount += cCount - bCount;
    c.data.valueCount -= bCount - brCount;
});
/**
 * Finds a node in the subtree rooted at a given node. It does this by calling a
 * special compare function that takes the current node as input and continues
 * depending on the output of this function:
 *
 * If the compare function returns a negative number on a given node, the search
 * will continue on the left subtree of the node.
 *
 * If the compare function returns a positive number on a given node, the search
 * will continue on the right subtree of the node.
 *
 * If the compare function returns zero on the node, then the current node will
 * be returned by `find`.
 *
 * @example
 * ```ts
 * // Assumes we have some compare function for the type of `value`.
 * const findByValue = (value, root) => find(root, (node) => {
 *   return compare(value, node.data.value)
 * })
 * ```
 *
 * @param node The root node of the tree to search.
 * @param progressUpdater The special compare function.
 */
function find(node, progressUpdater) {
    while (node !== undefined) {
        const next = progressUpdater(node);
        if (next < 0) {
            node = node.left;
        }
        else if (next > 0) {
            node = node.right;
        }
        else {
            break;
        }
    }
    return node;
}
/**
 * Inserts the value into the tree. The value must not be in the tree already.
 */
function insert(value, node, compare) {
    if (node === undefined) {
        return { data: { value, size: 1, valueCount: 1 } };
    }
    node.data.size += 1;
    const comparison = compare(value, node.data.value);
    if (comparison < 0) {
        node.left = insert(value, node.left, compare);
        return balanceRight(node);
    }
    else if (comparison > 0) {
        node.right = insert(value, node.right, compare);
        return balanceLeft(node);
    }
    return node;
}
/**
 * Removes a value from the subtree rooted at `node` and returns the root of the
 * resulting tree after deletion. The node must be in the tree already.
 *
 * @param value The value to remove.
 * @param node The root of the tree to remove the value from.
 * @param compare A comparison function to know which subtree to recurse into.
 */
function remove(value, node, compare) {
    if (node === undefined) {
        return undefined;
    }
    node.data.size -= 1;
    const comparison = compare(value, node.data.value);
    if (comparison < 0) {
        // The value is in the left subtree and is definitely in the tree.
        // eslint-disable-next-line @typescript-eslint/no-non-null-assertion
        node.left = remove(value, node.left, compare);
        return balanceLeft(node);
    }
    else if (comparison > 0) {
        // The value is in the right subtree and is definitely in the tree.
        // eslint-disable-next-line @typescript-eslint/no-non-null-assertion
        node.right = remove(value, node.right, compare);
        return balanceRight(node);
    }
    // The current node should be deleted.
    if (node.left !== undefined && node.right !== undefined) {
        // Two children. We should replace the current node with its predecessor
        // value and remove that value from the left subtree.
        // since it is a predecessor it will not have two children and we will not
        // recurse forever.
        let pre = node.left;
        while (pre.right !== undefined) {
            pre = pre.right;
        }
        node.data.value = pre.data.value;
        node.data.valueCount = pre.data.valueCount;
        node.left = remove(pre.data.value, node.left, compare);
        return balanceLeft(node);
    }
    // If at most one of the children is defined, they must be a single node, as
    // otherwise we would not be weight balanced.
    return node.left || node.right;
}

/**
 * The default comparison function. Designed to be consistend with
 * `Array.prototype.sort()`.`
 */
function defaultCmp(a, b) {
    const aStr = new String(a);
    const bStr = new String(b);
    return aStr < bStr ? -1 : aStr > bStr ? 1 : 0;
}
/**
 * A sorted array. Stays sorted through insertion and deletion.
 */
class SortedArray {
    /**
     * Constructs a new Sorted Array.
     *
     * @param compare The comparison function for the sorting. Compares two
     * elements a and b. If a is less than b, returns a negative number. If a is
     * greater than b returns a positive number. If a is equal to b returns zero.
     */
    constructor(compare = defaultCmp) {
        this[Symbol.toStringTag] = 'SortedArray';
        this.compare = compare;
    }
    findAtIndex(idx, callback) {
        return find(this.root, (node) => {
            var _a, _b, _c, _d;
            // The number of indicies less than the current node
            const ltSize = (_b = (_a = node.left) === null || _a === void 0 ? void 0 : _a.data.valueCount) !== null && _b !== void 0 ? _b : 0;
            // The number of indicies less than or equal to the current node
            const leSize = node.data.valueCount - ((_d = (_c = node.right) === null || _c === void 0 ? void 0 : _c.data.valueCount) !== null && _d !== void 0 ? _d : 0);
            if (callback !== undefined) {
                callback(node);
            }
            if (idx < ltSize) {
                // The index is in the left subtree
                return -1;
            }
            else if (idx >= leSize) {
                // the index is in the right subtree
                idx -= leSize;
                return 1;
            }
            return 0;
        });
    }
    /**
     * Removes the last value from the array and returns it. Optionally an index
     * to pop from can be specified.
     *
     * @param index The target index to pop from the array. Defaults to the last
     * element in the array. Must be within [0, arr.length)
     */
    pop(index) {
        if (this.root === undefined) {
            return undefined;
        }
        // We want to remove the rightmost node from the tree.
        const targetIndex = index !== null && index !== void 0 ? index : this.root.data.valueCount - 1;
        if (targetIndex < 0 || targetIndex >= this.length) {
            return undefined;
        }
        // The index is good, so we will definitely find a node.
        // eslint-disable-next-line @typescript-eslint/no-non-null-assertion
        const target = this.findAtIndex(targetIndex, (node) => {
            node.data.valueCount -= 1;
        });
        // The value that we are popping from the tree
        const result = target.data.value;
        if (target.data.valueCount === 0) {
            // We should delete the entire node.
            this.root = remove(result, this.root, this.compare);
        }
        return result;
    }
    /**
     * Inserts a value into the sorted array in sorted order.
     *
     * Note that if the comparison function returns zero on a value already in the
     * array, the inserted value will be "absorbed" into that value.
     *
     * For example:
     *
     * ```ts
     * const arr = SortedArray((a, b) => 0)
     *
     * arr.insert(5) // [ 5 ]
     * arr.insert(1000) // [ 5, 5 ]
     * ```
     *
     * The comparator returned 0 when called on 5 and 1000, so the 1000 value was
     * absorbed into the value 5.
     */
    insert(value) {
        const node = find(this.root, (node) => {
            node.data.valueCount += 1;
            return this.compare(value, node.data.value);
        });
        if (node === undefined) {
            this.root = insert(value, this.root, this.compare);
        }
        return this;
    }
    indexRange(value) {
        let left = 0;
        let right = 0;
        find(this.root, (node) => {
            var _a, _b, _c, _d;
            const next = this.compare(value, node.data.value);
            if (next >= 0) {
                right += node.data.valueCount - ((_b = (_a = node.right) === null || _a === void 0 ? void 0 : _a.data.valueCount) !== null && _b !== void 0 ? _b : 0);
            }
            if (next > 0) {
                left += node.data.valueCount - ((_d = (_c = node.right) === null || _c === void 0 ? void 0 : _c.data.valueCount) !== null && _d !== void 0 ? _d : 0);
            }
            return next;
        });
        return [left, right];
    }
    /**
     * Returns the largest index that the value could be inserted into.
     *
     * That is, `i`, such that for the sorted array `arr`, if `j` < `i`, `arr[j]`
     * ≤ `value` and if `j` ≥ `i`, `arr[j]` > `value`.
     */
    bisectRight(value) {
        return this.indexRange(value)[1];
    }
    /**
     * Similar to [[bisectRight]], but returns the smallest index that the value
     * could be inserted to.
     */
    bisectLeft(value) {
        return this.indexRange(value)[0];
    }
    /**
     * Returns the first index at which a given element can be found in the array,
     * or -1 if it is not present.
     *
     * @param value The value to locate in the array.
     * @param fromIdx The index to start the search at. If the index is greater
     * than or equal to the array's length, -1 is returned, which means the array
     * will not be searched. If the provided index value is a negative number, it
     * is taken as the offset from the end of the array. Note: if the provided
     * index is negative, the array is still searched from front to back. If the
     * provided index is 0, then the whole array will be searched. Default: 0
     * (entire array is searched).
     * @returns The first index of the element in the array; -1 if not found.
     */
    indexOf(value, fromIdx = 0) {
        if (fromIdx >= this.length) {
            return -1;
        }
        const [start, end] = this.indexRange(value);
        if (fromIdx <= start) {
            return start;
        }
        if (fromIdx < end) {
            return fromIdx;
        }
        return -1;
    }
    /**
     * Returns the last index at which a given element can be found in the array,
     * or -1 if it is not present. The array is searched backwards, starting at
     * fromIndex.
     *
     * @param value The value to locate in the array.
     * @param fromIndex The index at which to start searching backwards. Defaults
     * to the array's length minus one (arr.length - 1), i.e. the whole array will
     * be searched. If the index is greater than or equal to the length of the
     * array, the whole array will be searched. If negative, it is taken as the
     * offset from the end of the array. Note that even when the index is
     * negative, the array is still searched from back to front. If the calculated
     * index is less than 0, -1 is returned, i.e. the array will not be searched.
     * @returns The last index of the element in the array; -1 if not found.
     */
    lastIndexOf(value, fromIndex = -1) {
        if (fromIndex < 0) {
            if (fromIndex < -this.length) {
                return -1;
            }
            fromIndex += this.length;
        }
        const [start, end] = this.indexRange(value);
        if (fromIndex >= end) {
            return end - 1;
        }
        if (fromIndex >= start) {
            return fromIndex;
        }
        return -1;
    }
    /**
     * Executes a provided function once for each array element.
     *
     * @param callbackfn The function to execute on each element.
     * @param thisArg The value to use as this when executing callback.
     */
    forEach(callbackfn, thisArg) {
        callbackfn = callbackfn.bind(thisArg);
        for (const [i, val] of enumerate(this)) {
            callbackfn(val, i, this);
        }
    }
    entries() {
        return enumerate(this);
    }
    keys() {
        var _a, _b;
        return range((_b = (_a = this.root) === null || _a === void 0 ? void 0 : _a.data.valueCount) !== null && _b !== void 0 ? _b : 0);
    }
    values() {
        return this[Symbol.iterator]();
    }
    /**
     * Returns the value at the given index in the sorted order. (0-indexed)
     *
     * If the index is invalid, returns undefined.
     *
     * @param idx The index to lookup
     */
    get(idx) {
        var _a, _b, _c;
        // Make sure we have a valid index
        if (idx < 0 || idx >= ((_b = (_a = this.root) === null || _a === void 0 ? void 0 : _a.data.valueCount) !== null && _b !== void 0 ? _b : 0) || isNaN(idx)) {
            return undefined;
        }
        return (_c = this.findAtIndex(idx)) === null || _c === void 0 ? void 0 : _c.data.value;
    }
    /**
     * The number of distinct elements in the set.
     */
    get length() {
        var _a, _b;
        return (_b = (_a = this.root) === null || _a === void 0 ? void 0 : _a.data.valueCount) !== null && _b !== void 0 ? _b : 0;
    }
    /**
     * Iterates over the values in the set in sorted order.
     */
    *[Symbol.iterator]() {
        // Performs a morris traversal of the tree.
        let curr = this.root;
        const valuesFrom = function* (node) {
            var _a, _b;
            const lCount = (_b = (_a = node.left) === null || _a === void 0 ? void 0 : _a.data.valueCount) !== null && _b !== void 0 ? _b : 0;
            // The right child may be an ancestor, due to the morris traversal.
            // If so the count will be greater than the nodes count.
            const rCount = node.right && node.right.data.valueCount <= node.data.valueCount
                ? node.right.data.valueCount
                : 0;
            const count = node.data.valueCount - (lCount + rCount);
            yield* repeat(node.data.value, count);
        };
        while (curr) {
            if (!curr.left) {
                yield* valuesFrom(curr);
                curr = curr.right;
            }
            else {
                let pre = curr.left;
                while (pre.right && pre.right !== curr) {
                    pre = pre.right;
                }
                if (!pre.right) {
                    pre.right = curr;
                    curr = curr.left;
                }
                else {
                    pre.right = undefined;
                    yield* valuesFrom(curr);
                    curr = curr.right;
                }
            }
        }
    }
    /**
     * Creates a sorted array from an iterable.
     *
     * @param iterable The itereable to turn into a sorted array.
     * @param compare The compare method for the resulting SortedArray.
     * @returns A new sorted array whose elements are from the passed iterable.
     */
    static from(iterable, compare) {
        const result = new SortedArray(compare);
        for (const value of iterable) {
            result.insert(value);
        }
        return result;
    }
}

export { SortedArray, defaultCmp };
